#ifndef HUMANUS_TOOL_BASE_H
#define HUMANUS_TOOL_BASE_H

#include "schema.h"
#include "agent/base.h"
#include "config.h"
#include "mcp_stdio_client.h"
#include "mcp_sse_client.h"
#include <string>

namespace humanus {

// Represents the result of a tool execution.
struct ToolResult {
    json output;
    json error;
    json system;

    ToolResult(const json& output, const json& error = {}, const json& system = {}) 
        : output(output), error(error), system(system) {}

    bool empty() const {
        return output.empty() && error.empty() && system.empty();
    }

    ToolResult operator+(const ToolResult& other) const {
        auto combined_field = [](const json& field, const json& other_field) {
            if (field.empty()) {
                return other_field;
            }
            if (other_field.empty()) {
                return field;
            }
            json result = json::array();
            if (field.is_array()) {
                result.insert(result.end(), field.begin(), field.end());
            } else  {
                result.push_back(field);
            }
            if (other_field.is_array()) {
                result.insert(result.end(), other_field.begin(), other_field.end());
            } else  {
                result.push_back(other_field);
            }
            return result;
        };

        return {
            combined_field(output, other.output),
            combined_field(error, other.error),
            combined_field(system, other.system)
        };
    }

    static std::string parse_json_content(const json& content) {
        if (content.is_string()) {
            return content.get<std::string>();
        } else {
            return content.dump(2);
        }
    }

    std::string to_string() const {
        return !error.empty() ? "Error: " + parse_json_content(error) : parse_json_content(output);
    }
};

// A ToolResult that represents a failure.
struct ToolError : ToolResult {
    ToolError(const json& error) : ToolResult({}, error) {}
};

struct BaseTool {
    std::string name;
    std::string description;
    json parameters;

    BaseTool(const std::string& name, const std::string& description, const json& parameters) :
        name(name), description(description), parameters(parameters) {}

    // Execute the tool with given parameters.
    ToolResult operator()(const json& arguments) {
        return execute(arguments);
    }

    // Execute the tool with given parameters.
    virtual ToolResult execute(const json& arguments) = 0;

    json to_param() const {
        return {
            {"type", "function"},
            {"function", {
                {"name", name},
                {"description", description},
                {"parameters", parameters}
            }}
        };
    }
};

// Execute the tool with given parameters.
struct BaseMCPTool : BaseTool {
    std::string name;
    std::string description;
    json parameters;

    std::unique_ptr<mcp::client> _client;

    BaseMCPTool(const std::string& name, const std::string& description, const json& parameters)
    : BaseTool(name, description, parameters) {
        try {
            // Load tool configuration from config file
            auto _config = Config::get_instance().mcp_tool().at(name);

            if (_config.type == "stdio") {
                std::string command = _config.command;
                if (!_config.args.empty()) {
                    for (const auto& arg : _config.args) {
                        command += " " + arg;
                    }
                }
                _client = std::make_unique<mcp::stdio_client>(command, _config.env_vars);
            } else if (_config.type == "sse") {
                if (!_config.host.empty() && _config.port > 0) {
                    _client = std::make_unique<mcp::sse_client>(_config.host, _config.port);
                } else if (!_config.url.empty()) {
                    _client = std::make_unique<mcp::sse_client>(_config.url, "/sse");
                } else {
                    throw std::runtime_error("MCP SSE configuration missing host or port or url");
                }
            }

            _client->initialize(name + "_client", "0.0.1");
        } catch (const std::exception& e) {
            throw std::runtime_error("Failed to initialize MCP tool client for `" + name + "`: " + std::string(e.what()));
        }
    }
    
    ToolResult execute(const json& arguments) override {
        try {
            if (!_client) {
                throw std::runtime_error("MCP client not initialized");
            }
            json result = _client->call_tool(name, arguments);
            bool is_error = result.value("isError", false);
            // Return different ToolResult based on whether there is an error
            if (is_error) {
                return ToolError(result.value("content", json::array()));
            } else {
                return ToolResult(result.value("content", json::array()));
            }
        } catch (const std::exception& e) {
            return ToolError(e.what());
        }
    }
};

struct AgentAware : BaseTool {
    std::shared_ptr<BaseAgent> agent = nullptr;
};

}

#endif // HUMANUS_TOOL_BASE_H
